# from clove import constants
# # from clove.network.base import BaseNetwork
# # from clove.utils.external_source import get_transaction
# # networks = constants.NETWORKS_WITH_API
# #
# # print(networks)
# #
# # def audit_contract(network, txid=None):
# #     print(get_transaction(network=network, tx_hash=txid))
# #
# #
# # audit_contract('btc', '5bce375f07f6f373d18ec343b16375fd2accac3b7bde57a1c99f64f93cad8538')